@@ -6,7 +6,8 @@
  * Date        : 2008-09-25
  * Description : a tool to convert RAW file to DNG
  *
- * Copyright (C) 2008-2009 by Gilles Caulier <caulier dot gilles at gmail dot com>
+ * Copyright (C) 2008-2010 by Gilles Caulier <caulier dot gilles at gmail dot com>
+ * Copyright (C) 2010 by Jens Mueller <tschenser at gmx dot de>
  *
  * This program is free software; you can redistribute it
  * and/or modify it under the terms of the GNU General
@@ -60,7 +61,10 @@ extern "C"
 // KDE includes
 
 #include <kdebug.h>
-#include <kzip.h>
+
+// LibKDcraw includes
+
+#include <libkdcraw/version.h>
 
 // Libkexiv2 includes
 
@@ -70,13 +74,34 @@ extern "C"
 
 #include "dngwriter_p.h"
 #include "dngwriterhost.h"
-#include "dngwriter.h"
+
+#define CHUNK 65536
 
 using namespace KExiv2Iface;
 
 namespace DNGIface
 {
 
+enum DNGBayerPattern
+{
+    None = 1,
+    Standard,
+    Fuji,
+    FourColor
+};
+
+dng_date_time dngDateTime(QDateTime qDT)
+{
+    dng_date_time dngDT;
+    dngDT.fYear   = qDT.date().year();
+    dngDT.fMonth  = qDT.date().month();
+    dngDT.fDay    = qDT.date().day();
+    dngDT.fHour   = qDT.time().hour();
+    dngDT.fMinute = qDT.time().minute();
+    dngDT.fSecond = qDT.time().second();
+    return dngDT;
+}
+
 DNGWriter::DNGWriter()
          : d(new DNGWriterPrivate)
 {
@@ -180,13 +205,47 @@ int DNGWriter::convert()
         QFileInfo          outputInfo(dngFilePath);
         QByteArray         rawData;
         DcrawInfoContainer identify;
+        DcrawInfoContainer identifyMake;
 
         // -----------------------------------------------------------------------------------------
 
         kDebug() << "DNGWriter: Loading RAW data from " << inputInfo.fileName() ;
 
         KDcraw rawProcessor;
-        if (!rawProcessor.extractRAWData(inputFile(), rawData, identify))
+        if (!rawProcessor.rawFileIdentify(identifyMake, inputFile()))
+        {
+            kDebug() << "DNGWriter: Reading RAW file failed. Aborted..." ;
+            return -1;
+        }
+
+        dng_rect activeArea;
+        // TODO: need to get correct default crop size to avoid artifacts at the borders
+        int activeWidth  = 0;
+        int activeHeight = 0;
+        int outputHeight = 0;
+        int outputWidth  = 0;
+        if ((identifyMake.orientation == 5) || (identifyMake.orientation == 6))
+        {
+            outputHeight = identifyMake.outputSize.width();
+            outputWidth  = identifyMake.outputSize.height();
+        }
+        else
+        {
+            outputHeight = identifyMake.outputSize.height();
+            outputWidth  = identifyMake.outputSize.width();
+        }
+
+        bool useFullSensorImage = false;
+#if KDCRAW_VERSION >= 0x010300
+        // disable general fullsensor image see #240750
+        // seems this bug is fixed with libRaw 0.12beta4
+        if (identifyMake.make == "Canon")
+        {
+            useFullSensorImage = true;
+        }
+#endif
+
+        if (!rawProcessor.extractRAWData(inputFile(), rawData, identify, useFullSensorImage, 0))
         {
             kDebug() << "DNGWriter: Loading RAW data failed. Aborted..." ;
             return -1;
@@ -194,19 +253,19 @@ int DNGWriter::convert()
 
         if (d->cancel) return -2;
 
-        int width      = identify.imageSize.width();
-        int height     = identify.imageSize.height();
-        int pixelRange = 16;
-
         kDebug() << "DNGWriter: Raw data loaded:" ;
         kDebug() << "--- Data Size:     " << rawData.size() << " bytes" ;
         kDebug() << "--- Date:          " << identify.dateTime.toString(Qt::ISODate) ;
         kDebug() << "--- Make:          " << identify.make ;
         kDebug() << "--- Model:         " << identify.model ;
-        kDebug() << "--- Size:          " << width << "x" << height ;
+        kDebug() << "--- ImageSize:     " << identify.imageSize.width() << "x" << identify.imageSize.height() ;
+        kDebug() << "--- FullSize:      " << identify.fullSize.width() << "x" << identify.fullSize.height() ;
+        kDebug() << "--- OutputSize:    " << identify.outputSize.width() << "x" << identify.outputSize.height() ;
         kDebug() << "--- Orientation:   " << identify.orientation ;
         kDebug() << "--- Top margin:    " << identify.topMargin ;
         kDebug() << "--- Left margin:   " << identify.leftMargin ;
+        kDebug() << "--- Right margin:  " << identify.rightMargin ;
+        kDebug() << "--- Bottom margin: " << identify.bottomMargin ;
         kDebug() << "--- Filter:        " << identify.filterPattern ;
         kDebug() << "--- Colors:        " << identify.rawColors ;
         kDebug() << "--- Black:         " << identify.blackPoint ;
@@ -214,33 +273,89 @@ int DNGWriter::convert()
         kDebug() << "--- CAM->XYZ:" ;
 
         QString matrixVal;
-        for(int i=0; i<12; i+=3)
+        for(int i=0; i<4; i++)
         {
             kDebug() << "                   "
-                          << QString().sprintf("%03.4f  %03.4f  %03.4f", identify.cameraXYZMatrix[0][ i ],
-                                                                         identify.cameraXYZMatrix[0][i+1],
-                                                                         identify.cameraXYZMatrix[0][i+2])
-                          ;
+                     << QString().sprintf("%03.4f  %03.4f  %03.4f", identify.cameraXYZMatrix[i][0],
+                                                                    identify.cameraXYZMatrix[i][1],
+                                                                    identify.cameraXYZMatrix[i][2]);
         }
 
         // Check if CFA layout is supported by DNG SDK.
-        int bayerMosaic;
+        DNGBayerPattern bayerPattern = None;
+        uint32 filter = 0;
+        bool fujiRotate90 = false;
 
+        // Standard bayer layouts
         if (identify.filterPattern == QString("GRBGGRBGGRBGGRBG"))
         {
-            bayerMosaic = 0;
+            bayerPattern = Standard;
+            filter = 0;
         }
         else if (identify.filterPattern == QString("RGGBRGGBRGGBRGGB"))
         {
-            bayerMosaic = 1;
+            bayerPattern = Standard;
+            filter = 1;
         }
         else if (identify.filterPattern == QString("BGGRBGGRBGGRBGGR"))
         {
-            bayerMosaic = 2;
+            bayerPattern = Standard;
+            filter = 2;
         }
         else if (identify.filterPattern == QString("GBRGGBRGGBRGGBRG"))
         {
-            bayerMosaic = 3;
+            bayerPattern = Standard;
+            filter = 3;
+        }
+        // Fuji layouts
+        else if ((identify.filterPattern == QString("RGBGRGBGRGBGRGBG"))
+            && (identifyMake.make == QString("FUJIFILM")))
+        {
+            bayerPattern = Fuji;
+            fujiRotate90 = false;
+            filter = 0;
+        }
+        else if ((identify.filterPattern == QString("RBGGBRGGRBGGBRGG"))
+            && (identifyMake.make == QString("FUJIFILM")))
+        {
+            bayerPattern = Fuji;
+            fujiRotate90 = true;
+            filter = 0;
+        }
+        // Four color sensors
+        else if (identify.rawColors == 4)
+        {
+            bayerPattern = FourColor;
+            if (identify.filterPattern.length() != 16)
+            {
+                kDebug() << "DNGWriter: Bayer mosaic not supported. Aborted..." ;
+                return -1;
+            }
+            for (int i = 0; i < 16; i++)
+            {
+                filter = filter >> 2;
+                if (identify.filterPattern[i] == 'G')
+                {
+                    filter |= 0x00000000;
+                }
+                else if (identify.filterPattern[i] == 'M')
+                {
+                    filter |= 0x40000000;
+                }
+                else if (identify.filterPattern[i] == 'C')
+                {
+                    filter |= 0x80000000;
+                }
+                else if (identify.filterPattern[i] == 'Y')
+                {
+                    filter |= 0xC0000000;
+                }
+                else
+                {
+                    kDebug() << "DNGWriter: Bayer mosaic not supported. Aborted..." ;
+                    return -1;
+                }
+            }
         }
         else
         {
@@ -248,13 +363,45 @@ int DNGWriter::convert()
             return -1;
         }
 
+        if (true == fujiRotate90)
+        {
+            if (false == fujiRotate(rawData, identify))
+            {
+                kDebug() << "Can not rotate fuji image. Aborted...";
+                return -1;
+            }
+            int tmp = outputWidth;
+            outputWidth = outputHeight;
+            outputHeight = tmp;
+        }
+
+        if (true == useFullSensorImage)
+        {
+            activeArea   = dng_rect(identify.topMargin,
+                                    identify.leftMargin,
+                                    identify.outputSize.height() - identify.bottomMargin,
+                                    identify.outputSize.width() - identify.rightMargin);
+            activeWidth  = identify.outputSize.width() - identify.leftMargin - identify.rightMargin;
+            activeHeight = identify.outputSize.height() - identify.bottomMargin - identify.topMargin;
+        }
+        else
+        {
+            activeArea   = dng_rect(identify.outputSize.height(), identify.outputSize.width());
+            activeWidth  = identify.outputSize.width();
+            activeHeight = identify.outputSize.height();
+        }
+
         // Check if number of Raw Color components is supported.
-        if (identify.rawColors != 3)
+        if ((identify.rawColors != 3) && (identify.rawColors != 4))
         {
             kDebug() << "DNGWriter: Number of Raw color components not supported. Aborted..." ;
             return -1;
         }
 
+        int width      = identify.outputSize.width();
+        int height     = identify.outputSize.height();
+        int pixelRange = 16;
+
 /*      // NOTE: code to hack RAW data extraction
 
         QString   rawdataFilePath(inputInfo.completeBaseName() + QString(".dat"));
@@ -272,37 +419,22 @@ int DNGWriter::convert()
 */
         // -----------------------------------------------------------------------------------------
 
-        kDebug() << "DNGWriter: Formating RAW data to memory" ;
-
-        std::vector<unsigned short> raw_data;
-        raw_data.resize(rawData.size());
-        const unsigned short* dp = (const unsigned short*)rawData.data();
-        for (uint i = 0; i < raw_data.size()/2; i++)
-        {
-            raw_data[i] = *dp;
-            *dp++;
-        }
-
         if (d->cancel) return -2;
 
-        // -----------------------------------------------------------------------------------------
-
         kDebug() << "DNGWriter: DNG memory allocation and initialization" ;
 
         dng_memory_allocator memalloc(gDefaultDNGMemoryAllocator);
         dng_memory_stream stream(memalloc);
-        stream.Put(&raw_data.front(), raw_data.size()*sizeof(unsigned short));
+        stream.Put(rawData.data(), rawData.size());
 
         dng_rect rect(height, width);
         DNGWriterHost host(d, &memalloc);
 
-        // Unprocessed raw data.
-        host.SetKeepStage1(true);
+        host.SetSaveDNGVersion(dngVersion_SaveDefault);
+        host.SetSaveLinearDNG(false);
+        host.SetKeepOriginalFile(true);
 
-        // Linearized, tone curve processed data.
-        host.SetKeepStage2(true);
-
-        AutoPtr<dng_image> image(new dng_simple_image(rect, 1, ttShort, 1<<pixelRange, memalloc));
+        AutoPtr<dng_image> image(new dng_simple_image(rect, 1, ttShort, memalloc));
 
         if (d->cancel) return -2;
 
@@ -366,8 +498,8 @@ int DNGWriter::convert()
         ifd.fWhiteLevel[2]             = identify.whitePoint;
         ifd.fWhiteLevel[3]             = identify.whitePoint;
 
-        ifd.fDefaultScaleH             = dng_urational(1, 1);
-        ifd.fDefaultScaleV             = dng_urational(1, 1);
+        ifd.fDefaultScaleH             = dng_urational(outputWidth, activeWidth);
+        ifd.fDefaultScaleV             = dng_urational(outputHeight, activeHeight);
         ifd.fBestQualityScale          = dng_urational(1, 1);
 
         ifd.fCFARepeatPatternRows      = 0;
@@ -377,11 +509,11 @@ int DNGWriter::convert()
         ifd.fChromaBlurRadius          = dng_urational(0, 0);
         ifd.fAntiAliasStrength         = dng_urational(100, 100);
 
-        ifd.fActiveArea                = rect;
+        ifd.fActiveArea                = activeArea;
         ifd.fDefaultCropOriginH        = dng_urational(0, 1);
         ifd.fDefaultCropOriginV        = dng_urational(0, 1);
-        ifd.fDefaultCropSizeH          = dng_urational(width, 1);
-        ifd.fDefaultCropSizeV          = dng_urational(height, 1);
+        ifd.fDefaultCropSizeH          = dng_urational(activeWidth, 1);
+        ifd.fDefaultCropSizeV          = dng_urational(activeHeight, 1);
 
         ifd.fMaskedAreaCount           = 0;
         ifd.fLosslessJPEGBug16         = false;
@@ -406,9 +538,59 @@ int DNGWriter::convert()
         negative->SetLocalName(QString("%1 %2").arg(identify.make).arg(identify.model).toAscii());
         negative->SetOriginalRawFileName(inputInfo.fileName().toAscii());
 
-        negative->SetColorChannels(3);
-        negative->SetColorKeys(colorKeyRed, colorKeyGreen, colorKeyBlue);
-        negative->SetBayerMosaic(bayerMosaic);
+        negative->SetColorChannels(identify.rawColors);
+
+        ColorKeyCode colorCodes[4] = {colorKeyMaxEnum, colorKeyMaxEnum, colorKeyMaxEnum, colorKeyMaxEnum};
+        for(int i = 0; i < qMax(4, identify.colorKeys.length()); i++)
+        {
+            if (identify.colorKeys[i] == 'R')
+            {
+                colorCodes[i] = colorKeyRed;
+            }
+            else if (identify.colorKeys[i] == 'G')
+            {
+                colorCodes[i] = colorKeyGreen;
+            }
+            else if (identify.colorKeys[i] == 'B')
+            {
+                colorCodes[i] = colorKeyBlue;
+            }
+            else if (identify.colorKeys[i] == 'C')
+            {
+                colorCodes[i] = colorKeyCyan;
+            }
+            else if (identify.colorKeys[i] == 'M')
+            {
+                colorCodes[i] = colorKeyMagenta;
+            }
+            else if (identify.colorKeys[i] == 'Y')
+            {
+                colorCodes[i] = colorKeyYellow;
+            }
+        }
+
+        negative->SetColorKeys(colorCodes[0], colorCodes[1], colorCodes[2], colorCodes[3]);
+
+        switch (bayerPattern)
+        {
+        case Standard:
+            // Standard bayer mosaicing. All work fine there.
+            // Bayer CCD mask: http://en.wikipedia.org/wiki/Bayer_filter
+            negative->SetBayerMosaic(filter);
+            break;
+        case Fuji:
+            // TODO: Fuji is special case. Need to setup different bayer rules here.
+            // It do not work in all settings. Need indeep investiguations.
+            // Fuji superCCD: http://en.wikipedia.org/wiki/Super_CCD
+            negative->SetFujiMosaic(filter);
+            break;
+        case FourColor:
+            negative->SetQuadMosaic(filter);
+            break;
+        default:
+            kDebug() << "DNGWriter: unknown bayer pattern" ;
+            return -1;
+        }
 
         negative->SetWhiteLevel(identify.whitePoint, 0);
         negative->SetWhiteLevel(identify.whitePoint, 1);
@@ -428,6 +610,10 @@ int DNGWriter::convert()
                 orientation = dng_orientation::Rotate180();
                 break;
 
+            case DcrawInfoContainer::ORIENTATION_Mirror90CCW:
+                orientation = dng_orientation::Mirror90CCW();
+                break;
+
             case DcrawInfoContainer::ORIENTATION_90CCW:
                 orientation = dng_orientation::Rotate90CCW();
                 break;
@@ -450,39 +636,73 @@ int DNGWriter::convert()
 
         // -------------------------------------------------------------------------------
 
-        // Set Camera->XYZ Color matrix as profile.
-        dng_matrix_3by3 matrix(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);
-        dng_matrix_3by3 camXYZ;
 
         AutoPtr<dng_camera_profile> prof(new dng_camera_profile);
         prof->SetName(QString("%1 %2").arg(identify.make).arg(identify.model).toAscii());
 
-        camXYZ[0][0] = identify.cameraXYZMatrix[0][0];
-        camXYZ[0][1] = identify.cameraXYZMatrix[0][1];
-        camXYZ[0][2] = identify.cameraXYZMatrix[0][2];
-        camXYZ[1][0] = identify.cameraXYZMatrix[0][3];
-        camXYZ[1][1] = identify.cameraXYZMatrix[1][0];
-        camXYZ[1][2] = identify.cameraXYZMatrix[1][1];
-        camXYZ[2][0] = identify.cameraXYZMatrix[1][2];
-        camXYZ[2][1] = identify.cameraXYZMatrix[1][3];
-        camXYZ[2][2] = identify.cameraXYZMatrix[2][0];
+        // Set Camera->XYZ Color matrix as profile.
+        dng_matrix matrix;
+        switch (identify.rawColors)
+        {
+        case 3:
+            {
+                dng_matrix_3by3 camXYZ;
+                camXYZ[0][0] = identify.cameraXYZMatrix[0][0];
+                camXYZ[0][1] = identify.cameraXYZMatrix[0][1];
+                camXYZ[0][2] = identify.cameraXYZMatrix[0][2];
+                camXYZ[1][0] = identify.cameraXYZMatrix[1][0];
+                camXYZ[1][1] = identify.cameraXYZMatrix[1][1];
+                camXYZ[1][2] = identify.cameraXYZMatrix[1][2];
+                camXYZ[2][0] = identify.cameraXYZMatrix[2][0];
+                camXYZ[2][1] = identify.cameraXYZMatrix[2][1];
+                camXYZ[2][2] = identify.cameraXYZMatrix[2][2];
+                if (camXYZ.MaxEntry() == 0.0)
+                {
+                    kDebug() << "DNGWriter: Warning, camera XYZ Matrix is null" ;
+                    camXYZ = dng_matrix_3by3(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);
+                }
 
-        if (camXYZ.MaxEntry() == 0.0)
-            kDebug() << "DNGWriter: Warning, camera XYZ Matrix is null" ;
-        else
-            matrix = camXYZ;
+                matrix = camXYZ;
+
+                break;
+            }
+        case 4:
+            {
+                dng_matrix_4by3 camXYZ;
+                camXYZ[0][0] = identify.cameraXYZMatrix[0][0];
+                camXYZ[0][1] = identify.cameraXYZMatrix[0][1];
+                camXYZ[0][2] = identify.cameraXYZMatrix[0][2];
+                camXYZ[1][0] = identify.cameraXYZMatrix[1][0];
+                camXYZ[1][1] = identify.cameraXYZMatrix[1][1];
+                camXYZ[1][2] = identify.cameraXYZMatrix[1][2];
+                camXYZ[2][0] = identify.cameraXYZMatrix[2][0];
+                camXYZ[2][1] = identify.cameraXYZMatrix[2][1];
+                camXYZ[2][2] = identify.cameraXYZMatrix[2][2];
+                camXYZ[3][0] = identify.cameraXYZMatrix[3][0];
+                camXYZ[3][1] = identify.cameraXYZMatrix[3][1];
+                camXYZ[3][2] = identify.cameraXYZMatrix[3][2];
+                if (camXYZ.MaxEntry() == 0.0)
+                {
+                    kDebug() << "DNGWriter: Warning, camera XYZ Matrix is null" ;
+                    camXYZ = dng_matrix_4by3(0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);
+                }
+
+                matrix = camXYZ;
+
+                break;
+             }
+        }
 
         prof->SetColorMatrix1((dng_matrix) matrix);
         prof->SetCalibrationIlluminant1(lsD65);
         negative->AddProfile(prof);
 
-        // -------------------------------------------------------------------------------
-
-        // Clear "Camera WhiteXY"
-        negative->SetCameraWhiteXY(dng_xy_coord());
-
-        // This settings break color on preview and thumbnail
-        //negative->SetCameraNeutral(dng_vector_3(1.0, 1.0, 1.0));
+        dng_vector camNeutral(identify.rawColors);
+        for (int i = 0; i < identify.rawColors; i++)
+        {
+            camNeutral[i] = 1.0/identify.cameraMult[i];
+        }
+        negative->SetCameraNeutral(camNeutral);
 
         if (d->cancel) return -2;
 
@@ -494,29 +714,30 @@ int DNGWriter::convert()
         exif->fModel.Set_ASCII(identify.model.toAscii());
         exif->fMake.Set_ASCII(identify.make.toAscii());
 
-        // Time from original shot
-        dng_date_time dt;
-        dt.fYear   = identify.dateTime.date().year();
-        dt.fMonth  = identify.dateTime.date().month();
-        dt.fDay    = identify.dateTime.date().day();
-        dt.fHour   = identify.dateTime.time().hour();
-        dt.fMinute = identify.dateTime.time().minute();
-        dt.fSecond = identify.dateTime.time().second();
-
-        dng_date_time_info dti;
-        dti.SetDateTime(dt);
-        exif->fDateTimeOriginal  = dti;
-        exif->fDateTimeDigitized = dti;
-        negative->UpdateDateTime(dti);
-
         long int num, den;
         long     val;
         QString  str;
         KExiv2   meta;
         if (meta.load(inputFile()))
-        {
+	{
+	    // Time from original shot
+	    dng_date_time_info dti;
+	    dti.SetDateTime(dngDateTime(meta.getImageDateTime()));
+	    exif->fDateTimeOriginal = dti;
+
+	    dti.SetDateTime(dngDateTime(meta.getDigitizationDateTime(true)));
+	    exif->fDateTimeDigitized = dti;
+ 
+            negative->UpdateDateTime(dti);
+
             // String Tags
 
+            str = meta.getExifTagString("Exif.Image.Make");
+            if (!str.isEmpty()) exif->fMake.Set_ASCII(str.toAscii());
+
+            str = meta.getExifTagString("Exif.Image.Model");
+            if (!str.isEmpty()) exif->fModel.Set_ASCII(str.toAscii());
+
             str = meta.getExifTagString("Exif.Image.Software");
             if (!str.isEmpty()) exif->fSoftware.Set_ASCII(str.toAscii());
 
@@ -642,6 +863,137 @@ int DNGWriter::convert()
             if (meta.getExifTagLong("Exif.GPSInfo.GPSAltitudeRef", val))               exif->fGPSAltitudeRef           = (uint32)val;
             if (meta.getExifTagLong("Exif.GPSInfo.GPSDifferential", val))              exif->fGPSDifferential          = (uint32)val;
 
+            // Nikon Markernotes
+
+            if (meta.getExifTagRational("Exif.Nikon3.Lens", num, den, 0))              exif->fLensInfo[0]              = dng_urational(num, den);
+            if (meta.getExifTagRational("Exif.Nikon3.Lens", num, den, 1))              exif->fLensInfo[1]              = dng_urational(num, den);
+            if (meta.getExifTagRational("Exif.Nikon3.Lens", num, den, 2))              exif->fLensInfo[2]              = dng_urational(num, den);
+            if (meta.getExifTagRational("Exif.Nikon3.Lens", num, den, 3))              exif->fLensInfo[3]              = dng_urational(num, den);
+            if (meta.getExifTagRational("Exif.Nikon3.ISOSpeed", num, den, 1))          exif->fISOSpeedRatings[0]       = (uint32)num;
+
+            str = meta.getExifTagString("Exif.Nikon3.SerialNO");
+            if (!str.isEmpty()) str = str.replace("NO=", "");
+            if (!str.isEmpty()) str = str.replace(" ", "");
+            if (!str.isEmpty()) exif->fCameraSerialNumber.Set_ASCII(str.toAscii());
+
+            str = meta.getExifTagString("Exif.Nikon3.SerialNumber");
+            if (!str.isEmpty()) exif->fCameraSerialNumber.Set_ASCII(str.toAscii());
+
+            if (meta.getExifTagLong("Exif.Nikon3.ShutterCount", val))                  exif->fImageNumber              = (uint32)val;
+            if (meta.getExifTagLong("Exif.NikonLd1.LensIDNumber", val))                exif->fLensID.Set_ASCII((QString("%1").arg(val)).toAscii());
+            if (meta.getExifTagLong("Exif.NikonLd2.LensIDNumber", val))                exif->fLensID.Set_ASCII((QString("%1").arg(val)).toAscii());
+            if (meta.getExifTagLong("Exif.NikonLd3.LensIDNumber", val))                exif->fLensID.Set_ASCII((QString("%1").arg(val)).toAscii());
+            if (meta.getExifTagLong("Exif.NikonLd2.FocusDistance", val))               exif->fSubjectDistance          = dng_urational((uint32)pow(10.0, val/40.0), 100);
+            if (meta.getExifTagLong("Exif.NikonLd3.FocusDistance", val))               exif->fSubjectDistance          = dng_urational((uint32)pow(10.0, val/40.0), 100);
+            str = meta.getExifTagString("NikonLd1.LensIDNumber");
+            if (!str.isEmpty()) exif->fLensName.Set_ASCII(str.toAscii());
+            str = meta.getExifTagString("NikonLd2.LensIDNumber");
+            if (!str.isEmpty()) exif->fLensName.Set_ASCII(str.toAscii());
+            str = meta.getExifTagString("NikonLd3.LensIDNumber");
+            if (!str.isEmpty()) exif->fLensName.Set_ASCII(str.toAscii());
+
+            // Canon Markernotes
+
+            if (meta.getExifTagLong("Exif.Canon.SerialNumber", val))                   exif->fCameraSerialNumber.Set_ASCII((QString("%1").arg(val)).toAscii());
+            //if (meta.getExifTagLong("Exif.CanonCs.LensType", val))                     exif->fLensID.Set_ASCII((QString("%1").arg(val)).toAscii());
+            //if (meta.getExifTagLong("Exif.CanonCs.FlashActivity", val))                exif->fFlash                    = (uint32)val;
+            //if (meta.getExifTagLong("Exif.CanonFi.FileNumber", val))                   exif->fImageNumber              = (uint32)val;
+            //if (meta.getExifTagLong("Exif.CanonCs.MaxAperture", val))                  exif->fMaxApertureValue         = dng_urational(val, 100000);
+            if (meta.getExifTagLong("Exif.CanonCs.ExposureProgram", val))
+            {
+                switch (val)
+                {
+                case 1:
+                    exif->fExposureProgram = 2;
+                    break;
+                case 2:
+                    exif->fExposureProgram = 4;
+                    break;
+                case 3:
+                    exif->fExposureProgram = 3;
+                    break;
+                case 4:
+                    exif->fExposureProgram = 1;
+                    break;
+                default:
+                    break;
+                }
+            }
+            if (meta.getExifTagLong("Exif.CanonCs.MeteringMode", val))
+            {
+                switch (val)
+                {
+                case 1:
+                    exif->fMeteringMode = 3;
+                    break;
+                case 2:
+                    exif->fMeteringMode = 1;
+                    break;
+                case 3:
+                    exif->fMeteringMode = 5;
+                    break;
+                case 4:
+                    exif->fMeteringMode = 6;
+                    break;
+                case 5:
+                    exif->fMeteringMode = 2;
+                    break;
+                default:
+                    break;
+                }
+            }
+
+            long canonLensUnits = 1;
+            if (meta.getExifTagRational("Exif.CanonCs.Lens", num, den, 2))             canonLensUnits                  = num;
+            if (meta.getExifTagRational("Exif.CanonCs.Lens", num, den, 0))             exif->fLensInfo[1]              = dng_urational(num, canonLensUnits);
+            if (meta.getExifTagRational("Exif.CanonCs.Lens", num, den, 1))             exif->fLensInfo[0]              = dng_urational(num, canonLensUnits);
+            if (meta.getExifTagRational("Exif.Canon.FocalLength", num, den, 1))        exif->fFocalLength              = dng_urational(num, canonLensUnits);
+            long canonLensType = 65535;
+            if (meta.getExifTagLong("Exif.CanonCs.LensType", canonLensType))           exif->fLensID.Set_ASCII((QString("%1").arg(canonLensType)).toAscii());
+            str = meta.getExifTagString("Exif.Canon.LensModel");
+            if (!str.isEmpty())
+                exif->fLensName.Set_ASCII(str.toAscii());
+            else if (canonLensType != 65535)
+            {
+                str = meta.getExifTagString("Exif.CanonCs.LensType");
+                if (!str.isEmpty()) exif->fLensName.Set_ASCII(str.toAscii());
+            }
+            else if ((exif->fLensInfo[0].n > 0) && (exif->fLensInfo[1].n > 0))
+            {
+                double focalLenght = (double)exif->fLensInfo[0].n / (double)exif->fLensInfo[0].d;
+                QString lensName = QString("%1").arg(focalLenght, 0, 'f', 1);
+                if (exif->fLensInfo[0].n != exif->fLensInfo[1].n)
+                {
+                    focalLenght = (double)exif->fLensInfo[1].n / (double)exif->fLensInfo[1].d;
+                    lensName.append(QString("-%1").arg(focalLenght, 0, 'f', 1));
+                }
+                lensName.append(" mm");
+                exif->fLensName.Set_ASCII(lensName.toAscii());
+            }
+
+            str = meta.getExifTagString("Exif.Canon.OwnerName");
+            if (!str.isEmpty()) exif->fOwnerName.Set_ASCII(str.toAscii());
+
+            str = meta.getExifTagString("Exif.Canon.FirmwareVersion");
+            if (!str.isEmpty()) str = str.replace("Firmware", "");
+            if (!str.isEmpty()) str = str.replace("Version", "");
+            if (!str.isEmpty()) str = str.replace(" ", "");
+            if (!str.isEmpty()) exif->fFirmware.Set_ASCII(str.toAscii());
+
+            str = meta.getExifTagString("Exif.CanonSi.ISOSpeed");
+            if (!str.isEmpty()) exif->fISOSpeedRatings[0] = str.toInt();
+
+            // Pentax Markernotes
+
+            if (meta.getExifTagRational("Exif.Pentax.LensType", num, den, 0))          str = QString("%1").arg(num);
+            if (meta.getExifTagRational("Exif.Pentax.LensType", num, den, 1))          str = str + QString(" %1").arg(num);
+            if (!str.isEmpty()) exif->fLensID.Set_ASCII(str.toAscii());
+
+            // Olympus Markernotes
+
+            str = meta.getExifTagString("Exif.OlympusEq.SerialNumber");
+            if (!str.isEmpty()) exif->fCameraSerialNumber.Set_ASCII(str.toAscii());
+
             // Markernote backup.
 
             QByteArray mkrnts = meta.getExifTagData("Exif.Photo.MakerNote");
@@ -655,54 +1007,131 @@ int DNGWriter::convert()
                 AutoPtr<dng_memory_block> block(host.Allocate(mkrnts.size()));
                 stream.SetReadPosition(0);
                 stream.Get(block->Buffer(), mkrnts.size());
-                negative->SetMakerNote(block);
-                negative->SetMakerNoteSafety(true);
+
+                if (identifyMake.make != "Canon")
+                {
+                    negative->SetMakerNote(block);
+                    negative->SetMakerNoteSafety(true);
+                }
+
+                long mknOffset       = 0;
+                QString mknByteOrder = meta.getExifTagString("Exif.MakerNote.ByteOrder");
+
+                if ((meta.getExifTagLong("Exif.MakerNote.Offset", mknOffset)) && !mknByteOrder.isEmpty())
+                {
+                    dng_memory_stream streamPriv(memalloc);
+                    QByteArray identPriv;
+
+                    identPriv.append(QString("Adobe").toAscii());
+                    identPriv.append((char)0x00);
+                    identPriv.append(QString("MakN").toAscii());
+
+                    // next four byte are makernote size
+                    int mknSize = mkrnts.size();
+                    int size    = mknSize + 6;
+                    identPriv.append((char)((size >> 24) & 0xFF));
+                    identPriv.append((char)((size >> 16) & 0xFF));
+                    identPriv.append((char)((size >> 8 ) & 0xFF));
+                    identPriv.append((char)((size      ) & 0xFF));
+
+                    kDebug() << mknSize << " " << mknOffset << " " << mknByteOrder;
+
+                    // byte order
+                    identPriv.append(mknByteOrder.toAscii());
+
+                    // next four byte are original offset
+                    identPriv.append((char)((mknOffset >> 24) & 0xFF));
+                    identPriv.append((char)((mknOffset >> 16) & 0xFF));
+                    identPriv.append((char)((mknOffset >> 8 ) & 0xFF));
+                    identPriv.append((char)((mknOffset      ) & 0xFF));
+
+
+                    streamPriv.Put(identPriv, identPriv.size());
+                    streamPriv.Put(mkrnts.data(), mkrnts.size());
+                    AutoPtr<dng_memory_block> blockPriv(host.Allocate(identPriv.size() + mkrnts.size()));
+                    streamPriv.SetReadPosition(0);
+                    streamPriv.Get(blockPriv->Buffer(), identPriv.size() + mkrnts.size());
+                    negative->SetPrivateData(blockPriv);
+                }
             }
         }
 
         if (d->backupOriginalRawFile)
         {
-            kDebug() << "DNGWriter: Backup Original RAW file (" << inputInfo.size() << " bytes)" ;
+            kDebug() << "DNGWriter: Backup Original RAW file (" << inputInfo.size() << " bytes)";
 
-            // Compress Raw file data to Zip archive.
+            QFileInfo originalFileInfo(inputFile());
 
-            QTemporaryFile zipFile;
-            if (!zipFile.open())
+            QFile originalFile(originalFileInfo.absoluteFilePath());
+            originalFile.open(QIODevice::ReadOnly);
+            QDataStream originalDataStream(&originalFile);
+
+            quint32 forkLength = originalFileInfo.size();
+            quint32 forkBlocks = (quint32)floor((forkLength + 65535.0) / 65536.0);
+
+            QVector<quint32> offsets;
+            quint32 offset = (2 + forkBlocks) * sizeof(quint32);
+            offsets.push_back(offset);
+
+            QByteArray originalDataBlock;
+            originalDataBlock.resize(CHUNK);
+
+            QTemporaryFile compressedFile;
+            if (!compressedFile.open())
             {
-                kDebug() << "DNGWriter: Cannot open temporary file to write Zip Raw file. Aborted..." ;
+                kDebug() << "DNGWriter: Cannot open temporary file to write Zip Raw file. Aborted...";
                 return -1;
             }
-            KZip zipArchive(zipFile.fileName());
-            zipArchive.open(QIODevice::WriteOnly);
-            zipArchive.setCompression(KZip::DeflateCompression);
-            zipArchive.addLocalFile(inputFile(), inputFile());
-            zipArchive.close();
+            QDataStream compressedDataStream(&compressedFile);
 
-            // Load Zip Archive in a byte array
+            for (quint32 block = 0; block < forkBlocks; block++)
+            {
+                int originalBlockLength = originalDataStream.readRawData(originalDataBlock.data(), CHUNK);
 
-            QFileInfo zipFileInfo(zipFile.fileName());
-            QByteArray zipRawFileData;
-            zipRawFileData.resize(zipFileInfo.size());
-            QDataStream dataStream(&zipFile);
-            dataStream.readRawData(zipRawFileData.data(), zipRawFileData.size());
-            kDebug() << "DNGWriter: Zipped RAW file size " << zipRawFileData.size() << " bytes" ;
+                QByteArray compressedDataBlock = qCompress((const uchar*)originalDataBlock.data(), originalBlockLength, -1);
+                compressedDataBlock.remove(0, 4); // removes qCompress own header
+                kDebug() << "DNGWriter: compressed data block " << originalBlockLength << " -> " << compressedDataBlock.size();
 
-            // Pass byte array to DNG sdk and compute MD5 fingerprint.
+                offset += compressedDataBlock.size();
+                offsets.push_back(offset);
+
+                compressedDataStream.writeRawData(compressedDataBlock.data(), compressedDataBlock.size());
+            }
 
             dng_memory_allocator memalloc(gDefaultDNGMemoryAllocator);
-            dng_memory_stream stream(memalloc);
-            stream.Put(zipRawFileData.data(), zipRawFileData.size());
-            AutoPtr<dng_memory_block> block(host.Allocate(zipRawFileData.size()));
-            stream.SetReadPosition(0);
-            stream.Get(block->Buffer(), zipRawFileData.size());
+            dng_memory_stream tempDataStream(memalloc);
+            tempDataStream.SetBigEndian(true);
+            tempDataStream.Put_uint32(forkLength);
+            for (qint32 idx = 0; idx < offsets.size(); idx++)
+            {
+                tempDataStream.Put_uint32(offsets[idx]);
+            }
+
+            QByteArray compressedData;
+            compressedData.resize(compressedFile.size());
+            compressedFile.seek(0);
+            compressedDataStream.readRawData(compressedData.data(), compressedData.size());
+            tempDataStream.Put(compressedData.data(), compressedData.size());
+
+            compressedFile.remove();
+
+            tempDataStream.Put_uint32(0);
+            tempDataStream.Put_uint32(0);
+            tempDataStream.Put_uint32(0);
+            tempDataStream.Put_uint32(0);
+            tempDataStream.Put_uint32(0);
+            tempDataStream.Put_uint32(0);
+            tempDataStream.Put_uint32(0);
+
+            AutoPtr<dng_memory_block> block(host.Allocate(tempDataStream.Length()));
+            tempDataStream.SetReadPosition(0);
+            tempDataStream.Get(block->Buffer(), tempDataStream.Length());
 
             dng_md5_printer md5;
             md5.Process(block->Buffer(), block->LogicalSize());
             negative->SetOriginalRawFileData(block);
             negative->SetOriginalRawFileDigest(md5.Result());
             negative->ValidateOriginalRawFileDigest();
-
-            zipFile.remove();
         }
 
         if (d->cancel) return -2;
@@ -721,7 +1150,7 @@ int DNGWriter::convert()
         negative->BuildStage3Image(host);
 
         negative->SynchronizeMetadata();
-        negative->RebuildIPTC();
+        negative->RebuildIPTC(true, false);
 
         if (d->cancel) return -2;
 
@@ -875,4 +1304,27 @@ int DNGWriter::convert()
     return dng_error_none;
 }
 
+bool DNGWriter::fujiRotate(QByteArray& rawData, KDcrawIface::DcrawInfoContainer& identify)
+{
+    QByteArray tmpData(rawData);
+    int height = identify.outputSize.height();
+    int width  = identify.outputSize.width();
+
+    unsigned short* tmp    = (unsigned short*)tmpData.data();
+    unsigned short* output = (unsigned short*)rawData.data();
+
+    for (int row=0; row < height; row++)
+    {
+        for (int col=0; col < width; col++)
+        {
+            output[col * height + row] = tmp[row * width + col];
+        }
+    }
+
+    identify.orientation = DcrawInfoContainer::ORIENTATION_Mirror90CCW;
+    identify.outputSize  = QSize(height, width);
+    //TODO: rotate margins
+    return true;
+}
+
 }  // namespace DNGIface
