@@ -65,10 +65,6 @@
 #include <libkdcraw/version.h>
 #include <libkdcraw/kdcraw.h>
 
-#if KDCRAW_VERSION < 0x000400
-#include <libkdcraw/dcrawbinary.h>
-#endif
-
 // Local includes
 
 #include "pluginsversion.h"
@@ -127,8 +123,8 @@ FlickrTalker::~FlickrTalker()
 {
     if (m_job)
         m_job->kill();
-    if (m_photoSetsList)
-        delete m_photoSetsList;
+    delete m_photoSetsList;
+
 }
 
 /** Compute MD5 signature using url queries keys and values following Flickr notice:
@@ -600,11 +596,7 @@ bool FlickrTalker::addPhoto(const QString& photoPath, const FPhotoInfo& info,
     QImage image;
 
     // Check if RAW file.
-#if KDCRAW_VERSION < 0x000400
-    QString rawFilesExt(KDcrawIface::DcrawBinary::instance()->rawFiles());
-#else
     QString rawFilesExt(KDcrawIface::KDcraw::rawFiles());
-#endif
     QFileInfo fileInfo(photoPath);
     if (rawFilesExt.toUpper().contains(fileInfo.suffix().toUpper()))
         KDcrawIface::KDcraw::loadDcrawPreview(image, photoPath);
@@ -665,12 +657,12 @@ bool FlickrTalker::addPhoto(const QString& photoPath, const FPhotoInfo& info,
     return true;
 }
 
-QString FlickrTalker::getUserName()
+QString FlickrTalker::getUserName() const
 {
     return m_username;
 }
 
-QString FlickrTalker::getUserId()
+QString FlickrTalker::getUserId() const
 {
     return m_userId;
 }
